# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Higher-level In-App WAF Binding
"""
import json
import ctypes
import threading

from ._compat import UNICODE_CLASS
from ._ffi import PW_GOOD, PW_BLOCK, PW_MONITOR, PW_ERR_INVALID_CALL, \
        PW_ERR_TIMEOUT, pw_freeReturn, pw_initH, \
        pw_freeDiagnotics, pw_runH, pw_clearRuleH, pw_initAdditiveH, \
        pw_runAdditive, pw_clearAdditive
from .input import Input


class WAFReturn:
    """In-App WAF run return value."""

    def __init__(self, obj):
        self._obj = obj

    def __del__(self):
        if self._obj is None:
            return
        pw_freeReturn(self._obj)
        self._obj = None

    @property
    def error(self):
        """True if something went wrong during the run, the data field should
        be populated."""
        action = self.action
        return action is not None and action < PW_GOOD

    @property
    def block(self):
        """True if the request should be blocked."""
        return self.action == PW_BLOCK

    @property
    def report(self):
        """True if the request should be reported."""
        return self.action in (PW_BLOCK, PW_MONITOR)

    @property
    def timeout(self):
        """True if the In-App WAF timed out."""
        return self.action == PW_ERR_TIMEOUT

    @property
    def action(self):
        """Raw return value."""
        if self._obj is not None:
            return self._obj.action

    @property
    def data(self):
        """Raw data to report."""
        if self._obj is not None:
            return self._obj.data

    @property
    def perf_total_runtime(self):
        """WAF internal perf counter in micro seconds."""
        if self._obj is not None:
            return self._obj.perfTotalRuntime

    @property
    def perf_cache_hit_rate(self):
        """WAF internal cache hit counter."""
        if self._obj is not None:
            return self._obj.perfCacheHitRate

    @property
    def perf_data(self):
        """WAF internal perf data."""
        if self._obj is not None:
            return self._obj.perfData

    def __repr__(self):
        return "<WAFReturn block={0.block!r} report={0.report!r} error={0.error!r}>".format(self)


class WAFEngine:
    """
    In-App WAF engine interface.
    """

    def __init__(self, data):
        if isinstance(data, UNICODE_CLASS):
            data = data.encode("utf-8", errors="surrogatepass")
        diag_ptr = ctypes.c_char_p(None)
        try:
            self.handle = pw_initH(data, None, ctypes.byref(diag_ptr))
            if self.handle is None:
                if diag_ptr:
                    diag = diag_ptr.value.decode("utf-8")
                else:
                    diag = "unexpected error during WAF initialization"
                raise ValueError(diag)
        finally:
            if diag_ptr:
                pw_freeDiagnotics(diag_ptr)

    def create_context(self):
        """Start a new In-App WAF context for a request allowing to provide
        parameters progressively.
        """
        return WAFContext(self)

    def run(self, parameters, budget):
        """Run the In-App WAF on the given parameters."""
        if not isinstance(parameters, Input):
            parameters = Input.from_python(parameters)
        elif parameters.owned:
            raise ValueError("cannot use already submitted parameters")

        return WAFReturn(pw_runH(
            self.handle, parameters._obj, ctypes.c_size_t(budget)
        ))

    def __del__(self):
        handle = getattr(self, "handle", None)
        if handle is not None:
            pw_clearRuleH(handle)
            self.handle = None


class WAFContext:
    """In-App WAF request context."""

    def __init__(self, engine):
        # It is very important to keep a reference to the engine from the
        # context because it will prevent the garbage collertor to free
        # the engine while there are active contexts.
        self.engine = engine
        self.handle = pw_initAdditiveH(engine.handle)
        self.lock = threading.Lock()
        if self.handle is None:
            raise RuntimeError("unexpected error while creating a WAF context")

    def run(self, parameters, budget):
        """Run the In-App WAF with new parameters.
        """
        if not isinstance(parameters, Input):
            parameters = Input.from_python(parameters)
        elif parameters.owned:
            raise ValueError("cannot use already submitted parameters")

        with self.lock:
            ret = WAFReturn(pw_runAdditive(
                self.handle, parameters._obj, ctypes.c_size_t(budget)
            ))
            if ret.action not in (PW_ERR_INVALID_CALL, PW_ERR_TIMEOUT):
                parameters.owned = True
            return ret

    def __del__(self):
        if self.handle is not None:
            pw_clearAdditive(self.handle)
            self.handle = None



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
# 
#     https://www.sqreen.io/terms.html
# 
# /!\ This file is generated. DO NOT EDIT /!\
from ctypes import *

from ._ffi import _lib, _PWArgs, PW_LOG_FUNC

PWI_INVALID = 0
PWI_SIGNED_NUMBER = 1 << 0
PWI_UNSIGNED_NUMBER = 1 << 1
PWI_STRING = 1 << 2
PWI_ARRAY = 1 << 3
PWI_MAP = 1 << 4

PW_INPUT_TYPE = c_int

PWHandle = c_void_p

PWAddContext = c_void_p

PWArgs = _PWArgs

PWD_PARSING_JSON = 0
PWD_PARSING_RULE = 1
PWD_PARSING_RULE_FILTER = 2
PWD_OPERATOR_VALUE = 3
PWD_DUPLICATE_RULE = 4
PWD_PARSING_FLOW = 5
PWD_PARSING_FLOW_STEP = 6
PWD_MEANINGLESS_STEP = 7
PWD_DUPLICATE_FLOW = 8
PWD_DUPLICATE_FLOW_STEP = 9
PWD_STEP_HAS_INVALID_RULE = 10

PW_DIAG_CODE = c_int

PW_ERR_INTERNAL = -6
PW_ERR_TIMEOUT = -5
PW_ERR_INVALID_CALL = -4
PW_ERR_INVALID_RULE = -3
PW_ERR_INVALID_FLOW = -2
PW_ERR_NORULE = -1
PW_GOOD = 0
PW_MONITOR = 1
PW_BLOCK = 2

PW_RET_CODE = c_int

PWL_TRACE = 0
PWL_DEBUG = 1
PWL_INFO = 2
PWL_WARN = 3
PWL_ERROR = 4
_PWL_AFTER_LAST = 5

PW_LOG_LEVEL = c_int

class PWConfig(Structure):
    _fields_ = [
        ("maxArrayLength", c_uint64),
        ("maxMapDepth", c_uint64),
    ]

_pw_init = _lib.pw_init
_pw_init.restype = c_bool
_pw_init.argtypes = [c_char_p, c_char_p, POINTER(PWConfig), POINTER(c_char_p)]
def pw_init(ruleName, wafRule, config, errors):
    # type: (Any, Any, Any, Any) -> c_bool
    return _pw_init(ruleName, wafRule, config, errors)  # type: ignore

class PWRet(Structure):
    _fields_ = [
        ("action", PW_RET_CODE),
        ("data", c_char_p),
        ("perfData", c_char_p),
        ("perfTotalRuntime", c_uint32),
        ("perfCacheHitRate", c_uint32),
    ]

_pw_run = _lib.pw_run
_pw_run.restype = PWRet
_pw_run.argtypes = [c_char_p, PWArgs, c_uint64]
def pw_run(ruleName, parameters, timeLeftInUs):
    # type: (Any, Any, Any) -> PWRet
    return _pw_run(ruleName, parameters, timeLeftInUs)  # type: ignore

_pw_clearRule = _lib.pw_clearRule
_pw_clearRule.restype = None
_pw_clearRule.argtypes = [c_char_p]
def pw_clearRule(ruleName):
    # type: (Any) -> None
    return _pw_clearRule(ruleName)  # type: ignore

_pw_clearAll = _lib.pw_clearAll
_pw_clearAll.restype = None
_pw_clearAll.argtypes = []
def pw_clearAll():
    # type: () -> None
    return _pw_clearAll()  # type: ignore

_pw_initH = _lib.pw_initH
_pw_initH.restype = PWHandle
_pw_initH.argtypes = [c_char_p, POINTER(PWConfig), POINTER(c_char_p)]
def pw_initH(wafRule, config, errors):
    # type: (Any, Any, Any) -> PWHandle
    return _pw_initH(wafRule, config, errors)  # type: ignore

_pw_runH = _lib.pw_runH
_pw_runH.restype = PWRet
_pw_runH.argtypes = [PWHandle, PWArgs, c_uint64]
def pw_runH(wafHandle, parameters, timeLeftInUs):
    # type: (Any, Any, Any) -> PWRet
    return _pw_runH(wafHandle, parameters, timeLeftInUs)  # type: ignore

_pw_clearRuleH = _lib.pw_clearRuleH
_pw_clearRuleH.restype = None
_pw_clearRuleH.argtypes = [PWHandle]
def pw_clearRuleH(wafHandle):
    # type: (Any) -> None
    return _pw_clearRuleH(wafHandle)  # type: ignore

_pw_initAdditive = _lib.pw_initAdditive
_pw_initAdditive.restype = PWAddContext
_pw_initAdditive.argtypes = [c_char_p]
def pw_initAdditive(ruleName):
    # type: (Any) -> PWAddContext
    return _pw_initAdditive(ruleName)  # type: ignore

_pw_initAdditiveH = _lib.pw_initAdditiveH
_pw_initAdditiveH.restype = PWAddContext
_pw_initAdditiveH.argtypes = [PWHandle]
def pw_initAdditiveH(powerwafHandle):
    # type: (Any) -> PWAddContext
    return _pw_initAdditiveH(powerwafHandle)  # type: ignore

_pw_runAdditive = _lib.pw_runAdditive
_pw_runAdditive.restype = PWRet
_pw_runAdditive.argtypes = [PWAddContext, PWArgs, c_uint64]
def pw_runAdditive(context, newArgs, timeLeftInUs):
    # type: (Any, Any, Any) -> PWRet
    return _pw_runAdditive(context, newArgs, timeLeftInUs)  # type: ignore

_pw_clearAdditive = _lib.pw_clearAdditive
_pw_clearAdditive.restype = None
_pw_clearAdditive.argtypes = [PWAddContext]
def pw_clearAdditive(context):
    # type: (Any) -> None
    return _pw_clearAdditive(context)  # type: ignore

_pw_freeDiagnotics = _lib.pw_freeDiagnotics
_pw_freeDiagnotics.restype = None
_pw_freeDiagnotics.argtypes = [c_char_p]
def pw_freeDiagnotics(errors):
    # type: (Any) -> None
    return _pw_freeDiagnotics(errors)  # type: ignore

_pw_freeReturn = _lib.pw_freeReturn
_pw_freeReturn.restype = None
_pw_freeReturn.argtypes = [PWRet]
def pw_freeReturn(output):
    # type: (Any) -> None
    return _pw_freeReturn(output)  # type: ignore

class PWVersion(Structure):
    _fields_ = [
        ("major", c_uint16),
        ("minor", c_uint16),
        ("patch", c_uint16),
    ]

_pw_getVersion = _lib.pw_getVersion
_pw_getVersion.restype = PWVersion
_pw_getVersion.argtypes = []
def pw_getVersion():
    # type: () -> PWVersion
    return _pw_getVersion()  # type: ignore

_pw_setupLogging = _lib.pw_setupLogging
_pw_setupLogging.restype = c_bool
_pw_setupLogging.argtypes = [PW_LOG_FUNC, PW_LOG_LEVEL]
def pw_setupLogging(cb, min_level):
    # type: (Any, Any) -> c_bool
    return _pw_setupLogging(cb, min_level)  # type: ignore

_pw_getInvalid = _lib.pw_getInvalid
_pw_getInvalid.restype = PWArgs
_pw_getInvalid.argtypes = []
def pw_getInvalid():
    # type: () -> PWArgs
    return _pw_getInvalid()  # type: ignore

_pw_createStringWithLength = _lib.pw_createStringWithLength
_pw_createStringWithLength.restype = PWArgs
_pw_createStringWithLength.argtypes = [c_char_p, c_uint64]
def pw_createStringWithLength(string, length):
    # type: (Any, Any) -> PWArgs
    return _pw_createStringWithLength(string, length)  # type: ignore

_pw_createString = _lib.pw_createString
_pw_createString.restype = PWArgs
_pw_createString.argtypes = [c_char_p]
def pw_createString(string):
    # type: (Any) -> PWArgs
    return _pw_createString(string)  # type: ignore

_pw_createInt = _lib.pw_createInt
_pw_createInt.restype = PWArgs
_pw_createInt.argtypes = [c_int64]
def pw_createInt(value):
    # type: (Any) -> PWArgs
    return _pw_createInt(value)  # type: ignore

_pw_createUint = _lib.pw_createUint
_pw_createUint.restype = PWArgs
_pw_createUint.argtypes = [c_uint64]
def pw_createUint(value):
    # type: (Any) -> PWArgs
    return _pw_createUint(value)  # type: ignore

_pw_createArray = _lib.pw_createArray
_pw_createArray.restype = PWArgs
_pw_createArray.argtypes = []
def pw_createArray():
    # type: () -> PWArgs
    return _pw_createArray()  # type: ignore

_pw_createMap = _lib.pw_createMap
_pw_createMap.restype = PWArgs
_pw_createMap.argtypes = []
def pw_createMap():
    # type: () -> PWArgs
    return _pw_createMap()  # type: ignore

_pw_addArray = _lib.pw_addArray
_pw_addArray.restype = c_bool
_pw_addArray.argtypes = [POINTER(PWArgs), PWArgs]
def pw_addArray(array, entry):
    # type: (Any, Any) -> c_bool
    return _pw_addArray(array, entry)  # type: ignore

_pw_addMap = _lib.pw_addMap
_pw_addMap.restype = c_bool
_pw_addMap.argtypes = [POINTER(PWArgs), c_char_p, c_uint64, PWArgs]
def pw_addMap(map, entryName, entryNameLength, entry):
    # type: (Any, Any, Any, Any) -> c_bool
    return _pw_addMap(map, entryName, entryNameLength, entry)  # type: ignore

_pw_freeArg = _lib.pw_freeArg
_pw_freeArg.restype = None
_pw_freeArg.argtypes = [POINTER(PWArgs)]
def pw_freeArg(input):
    # type: (Any) -> None
    return _pw_freeArg(input)  # type: ignore

_pw_memAlloc = _lib.pw_memAlloc
_pw_memAlloc.restype = c_void_p
_pw_memAlloc.argtypes = [c_uint64]
def pw_memAlloc(size):
    # type: (Any) -> pointer
    return _pw_memAlloc(size)  # type: ignore

_pw_memRealloc = _lib.pw_memRealloc
_pw_memRealloc.restype = c_void_p
_pw_memRealloc.argtypes = [c_void_p, c_uint64]
def pw_memRealloc(ptr, size):
    # type: (Any, Any) -> pointer
    return _pw_memRealloc(ptr, size)  # type: ignore

_pw_memFree = _lib.pw_memFree
_pw_memFree.restype = None
_pw_memFree.argtypes = [c_void_p]
def pw_memFree(ptr):
    # type: (Any) -> None
    return _pw_memFree(ptr)  # type: ignore

_pw_initString = _lib.pw_initString
_pw_initString.restype = PWArgs
_pw_initString.argtypes = [c_char_p, c_uint64]
def pw_initString(string, length):
    # type: (Any, Any) -> PWArgs
    return _pw_initString(string, length)  # type: ignore

_pw_addMapNoCopy = _lib.pw_addMapNoCopy
_pw_addMapNoCopy.restype = c_bool
_pw_addMapNoCopy.argtypes = [POINTER(PWArgs), c_char_p, c_uint64, PWArgs]
def pw_addMapNoCopy(map, entryName, entryNameLength, entry):
    # type: (Any, Any, Any, Any) -> c_bool
    return _pw_addMapNoCopy(map, entryName, entryNameLength, entry)  # type: ignore



# Copyright (c) 2016, 2017, 2018, 2019, 2020, Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#

# The lib version is always contained at the beginning of __version__
__version__ = "1.0.4.0.0"
__lib_version__ = "1.0.4"



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Sqreen Python Native Module"""

from ._ffi import get_lib_version

__all__ = ["get_lib_version"]



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Sqreen Python Native Module"""
import ctypes
import logging
import os
import sys
import sysconfig

try:
    import pkg_resources
except ImportError:
    pkg_resources = None  # pragma: no cover

from . import __about__

LOGGER = logging.getLogger("sqreen.native.ffi")


def _get_libc_name():
    """Return the libc of the system."""
    target = sysconfig.get_config_var("HOST_GNU_TYPE")
    if target is not None and target.endswith("musl"):
        return "muslc"
    return "glibc"


def _get_lib_path(name):
    """Return the path of the library called `name`."""
    if os.name == "posix" and sys.platform == "darwin":
        prefix, ext = "lib", ".dylib"
    elif sys.platform == "win32":
        prefix, ext = "", ".dll"
    else:
        prefix, ext = "lib", ".{}.so".format(_get_libc_name())
    fn = None
    meipass = getattr(sys, "_MEIPASS", None)
    if meipass is not None:
        fn = os.path.join(meipass, prefix + name + ext)
    if fn is None and pkg_resources is not None:
        fn = pkg_resources.resource_filename("sq_native", prefix + name + ext)
    if fn is None:
        root_dir = os.path.dirname(os.path.abspath(__file__))
        fn = os.path.join(root_dir, prefix + name + ext)
    return fn


def _load_library(name):
    """Load a native library located in this module."""
    path = _get_lib_path(name)
    if path is None or not os.path.exists(path):
        raise RuntimeError("Native library not available at {}".format(path))
    return ctypes.cdll.LoadLibrary(path)


# defined manually because the value field is declared as a union.
# It must be kept up-to-date with the library header file.
class _PWArgs(ctypes.Structure):
    _fields_ = [
        ("parameterName", ctypes.POINTER(ctypes.c_char)),
        ("parameterNameLength", ctypes.c_uint64),
        ("value", ctypes.c_void_p),
        ("nbEntries", ctypes.c_uint64),
        ("type", ctypes.c_int),
    ]

PW_LOG_FUNC = ctypes.CFUNCTYPE(
    None, ctypes.c_int, ctypes.c_char_p, ctypes.c_char_p,
    ctypes.c_int, ctypes.c_char_p, ctypes.c_size_t)


# stage 1: load the library then import bindings
_lib = _load_library("Sqreen")

from ._bindings import *

# stage 2: check library version and setup logging if required
def get_lib_version():
    # type: () -> str
    """Get the library version as a string."""
    return "{0.major}.{0.minor}.{0.patch}".format(pw_getVersion())

_lib_version = get_lib_version()
if _lib_version != __about__.__lib_version__:
    raise RuntimeError("Native library version mismatch: {} != {}"
                       .format(_lib_version, __about__.__lib_version__))



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Binding for the WAF Input Data Structure
"""
import ctypes

from ._compat import UNICODE_CLASS
from ._ffi import pw_createArray, pw_createInt, pw_createUint, pw_createMap, \
        pw_createStringWithLength, pw_addArray, pw_addMap, pw_freeArg


def create_string(value, max_string_length=4096):
    if isinstance(value, UNICODE_CLASS):
        value = value[:max_string_length].encode("utf-8", errors="surrogatepass")

    if not isinstance(value, bytes):
        raise ValueError("value must be a string or bytes")

    value = value[:max_string_length]
    return pw_createStringWithLength(value, len(value))


def append_to_map(array, key, value):
    if isinstance(key, UNICODE_CLASS):
        key = key.encode("utf-8", errors="surrogatepass")

    if not isinstance(key, bytes):
        raise ValueError("value must be a string or bytes")

    return pw_addMap(ctypes.byref(array), key, 0, value)


def free(value):
    pw_freeArg(ctypes.byref(value))


def create(value, max_depth=10, ignore_none=True, max_string_length=4096,
           max_items=150):
    """ Lower-level function to convert a Python value to input value
    """
    if isinstance(value, str) or isinstance(value, bytes):
        return create_string(value, max_string_length=max_string_length)

    if isinstance(value, bool):
        return pw_createUint(ctypes.c_uint64(int(value)))

    if isinstance(value, int):
        if value < 0:
            return pw_createInt(ctypes.c_int64(value))
        else:
            return pw_createUint(ctypes.c_uint64(value))

    if isinstance(value, list) or isinstance(value, tuple):
        obj = pw_createArray()
        if max_depth <= 0:
            # ignore if deeply nested
            return obj
        for i, item in enumerate(value):
            if i >= max_items or (item is None and ignore_none):
                continue
            item_obj = create(item, max_depth=max_depth - 1)
            ret = pw_addArray(ctypes.byref(obj), item_obj)
            if ret is False:
                free(item_obj)
        return obj

    if isinstance(value, dict):
        obj = pw_createMap()
        if max_depth <= 0:
            # ignore if deeply nested
            return obj
        for i, (k, v) in enumerate(value.items()):
            if i >= max_items or (v is None and ignore_none):
                continue
            item_obj = create(v, max_depth=max_depth - 1)
            ret = append_to_map(obj, k, item_obj)
            if ret is False:
                free(item_obj)
        return obj

    return create_string(UNICODE_CLASS(value), max_string_length=max_string_length)


class Input:
    """
    Higher-level bridge between Python values and input values (PWArgs).
    """

    def __init__(self, obj):
        self._obj = obj
        self.owned = False

    def __del__(self):
        # If the PWArgs is owned, it must not be freed
        if self._obj is None or self.owned:
            return
        free(self._obj)
        self._obj = None

    @classmethod
    def from_python(cls, value, **kwargs):
        """ Convert a Python value to a managed input.
        """
        return cls(create(value, **kwargs))

    def __repr__(self):
        return "<{} obj={!r}>".format(self.__class__.__name__, self._obj)


# Alias for versions <1.0
PWArgs = Input

__all__ = ["Input"]



import io
import os

import setuptools

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
ABOUT_PATH = os.path.join(
    ROOT_DIR, "sq_native", "__about__.py")
ABOUT = {}  # type: ignore

# Read the version from the source
with io.open(ABOUT_PATH, encoding="utf-8") as f:
    exec(f.read(), ABOUT)

setuptools.setup(version=ABOUT["__version__"])
